/*
        OpenGL Framework by Thomas Baier (thomas.baier@stmuc.com)
*/
#pragma comment(lib, "opengl32.lib")
#pragma comment(lib, "glu32.lib")

#include <windows.h>
#include <GL\gl.h>
#include <GL\glu.h>
#include <math.h>
#include <stdio.h>

#define FULLSCREEN                              (0)
#define WIDTH                                   (640)
#define HEIGHT                                  (480)
#define BITDEPTH                                (16)
#define START_WITH_ANIMATION					(0)

/********************************************************************************/

#include "scene.h"

/*******************************************************************************/


/*******************************************************************************

        Main

*******************************************************************************/

/* image map structures */
struct DIB2D
{
        BITMAPINFOHEADER *Info;
        RGBQUAD *palette;
        BYTE    *bits;
};

struct GLTXTLOAD
{
        GLint format;
        GLint perpixel;
        GLint Width;
        GLint Height;
        BYTE* bits;
};


static char szAppName[] = "3DWin: "SCENE_NAME;
static GLsizei glnWidth, glnHeight;
static GLdouble gldAspect;

static int face_number = 0;

static HPALETTE hPalette        = NULL;
static GLfloat nSize            = 0.0f;
static GLfloat nCol             = 0.0f;
static GLfloat nTop             = 1.0f;
static GLfloat nBottom          = 0.0f;
static int bFlag                = 1;
static dFlag                    = 1;
static aFlag                    = START_WITH_ANIMATION; // animation

LONG WINAPI MainWndProc (HWND, UINT, WPARAM, LPARAM);
BOOL CreateGLWindow(char* title, int width, int height, int bits, bool fullscreenflag);
static void SetDCPixelFormat (HWND hwnd,HDC hdc);
static BOOL ChangeResolution(LONG w, LONG h, LONG bitdepth, int mode);
static void InitializeRC (void);
static void Resize(void);
static void DrawScene (HDC hdc, float nAngleX, float nAngleY, float nAngleZ, float nDistanceX, float nDistanceY, float nDistanceZ);
static void SelectMaterial(int i);
static void LoadTexture(char* filename);
static void SetupLights( void );

#define ROT_STEP        (1.0f)
#define MOVE_STEP       (0.1f)

// Display Lists
#define LIGHT_LIST		(1)
#define POLYGON_LIST	(2)

int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    LPSTR lpszCmdLine, int nCmdShow)
{
        WNDCLASSEX  wndclass;
        HWND hWnd;
        MSG msg;

        /* Register the frame class */
        wndclass.cbSize        = sizeof (wndclass);
        wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
        wndclass.lpfnWndProc   = (WNDPROC) MainWndProc;
        wndclass.cbClsExtra    = 0;
        wndclass.cbWndExtra    = 0;
        wndclass.hInstance     = hInstance;
        wndclass.hIcon         = 0;
        wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW);
        wndclass.hbrBackground = NULL;
        wndclass.lpszMenuName  = NULL;
        wndclass.lpszClassName = szAppName;
        wndclass.hIconSm       = 0;


        if (!RegisterClassEx(&wndclass))
                return FALSE;

        if ( FULLSCREEN )
        {
			if (!ChangeResolution(WIDTH,HEIGHT,BITDEPTH, 0))
				return 0;
        }

        if ( FULLSCREEN )
        {
			hWnd = CreateWindowEx(WS_EX_APPWINDOW /*| WS_EX_TOPMOST */,
									szAppName, szAppName,
									WS_POPUP |  WS_CLIPSIBLINGS,
									0, 0, WIDTH, HEIGHT,
									NULL, NULL, hInstance, NULL);
        }
        else
        {
			LONG  h;
			LONG  w;

			h = GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CXDLGFRAME)*2 + HEIGHT;
			w = GetSystemMetrics(SM_CXDLGFRAME)*2 + WIDTH;

			hWnd = CreateWindowEx(WS_EX_APPWINDOW,
									szAppName, szAppName,
									WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
									1, 1, w, h,
									NULL, NULL, hInstance, NULL);
        }


        /* Make sure window was created */
        if (!hWnd)
                return FALSE;

        ShowWindow (hWnd, nCmdShow);
        UpdateWindow (hWnd);

		char title[4096];
		sprintf(title, "%s (%d faces)", szAppName, face_number);
		SetWindowText(hWnd, title);

		while (GetMessage (&msg, NULL, 0, 0))
        {
			TranslateMessage (&msg);
			DispatchMessage (&msg);
		}
		return msg.wParam;
}


LONG WINAPI MainWndProc (HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static HDC hdc;
    static HGLRC hrc;
    PAINTSTRUCT ps;
    static float nAngleX = 0.0f;
    static float nAngleY = 0.0f;
    static float nAngleZ = 0.0f;
    static float nDistanceX = 0.0f;
    static float nDistanceY = 0.0f;
    static float nDistanceZ = -8.0f;
    static INT nTimer;
    int n;

    switch (msg)
	{

		case WM_CREATE:
        {
			hdc = GetDC (hwnd);
            SetDCPixelFormat (hwnd,hdc);
            hrc = wglCreateContext (hdc);
            wglMakeCurrent (hdc, hrc);
			SetForegroundWindow(hwnd);						// Slightly Higher Priority
            InitializeRC ();

            nTimer = SetTimer (hwnd, 1, 50, NULL);
            return 0;
        }
        break;

        case WM_SIZE:
        {
			glnWidth = (GLsizei) LOWORD (lParam);
            glnHeight = (GLsizei) HIWORD (lParam);
			Resize();
			return 0;
        }
        break;

        case WM_PAINT:
        {
			BeginPaint (hwnd, &ps);
            DrawScene (hdc, nAngleX, nAngleY, nAngleZ, nDistanceX, nDistanceY, nDistanceZ);
            EndPaint (hwnd, &ps);
            return 0;
		}
        break;

		case WM_TIMER:
        {
			if(aFlag == 0)
				break;

			nAngleZ += 1;
            if (nAngleZ >= 360)
				nAngleZ -= 360;

			if (bFlag == 1)
				nSize += 0.05f;

			nCol += 0.01f;
            if (nSize >= nTop)
				bFlag = 0;

			if (bFlag == 0)
				nSize -= 0.05f;
			nCol -= 0.01f;

			if (nSize <= nBottom)
				bFlag = 1;

			InvalidateRect (hwnd, NULL, FALSE);

			return 0;
		}
        break;

		case WM_QUERYNEWPALETTE:
        {
			if (hPalette != NULL)
            {
				if (n = RealizePalette (hdc))
					InvalidateRect (hwnd, NULL, FALSE);
				return n;
			}
		}
        break;

				case WM_PALETTECHANGED:
                {
					if ((hPalette != NULL) && ((HWND) wParam != hwnd))
                    {
                        if (RealizePalette (hdc))
							UpdateColors (hdc);
						return 0;
					}
                }
				break;

                case WM_DESTROY:
                {
					wglMakeCurrent (NULL, NULL);
                    wglDeleteContext (hrc);
					ReleaseDC (hwnd, hdc);
                    if (hPalette != NULL)
						DeleteObject (hPalette);
					KillTimer (hwnd, nTimer);
					if ( FULLSCREEN )
						ChangeResolution(WIDTH,HEIGHT,BITDEPTH, 1);
                    PostQuitMessage (0);
					return 0;
                }
                break;

                case WM_CHAR:
                {
					char chCharCode = (TCHAR) wParam;
					if(chCharCode == 0x1B)
					{
						SendMessage(hwnd, WM_DESTROY, 0,0);
                        return 0;
                        break;
					}

                        if(chCharCode == 'l')   // line draw on/off
                                dFlag ^= 1;
                        if(chCharCode == 'a')   // animation on/off
                                aFlag ^= 1;
                        if(chCharCode == 'r')   // reset
                        {
                                nSize   = 0.0f;
                                nCol    = 0.0f;
                                nTop    = 1.0f;
                                nBottom = 0.0f;
                                bFlag   = 1;
                                dFlag   = 1;
                                aFlag   = START_WITH_ANIMATION;
                                nAngleX = 0.0f;
                                nAngleY = 0.0f;
                                nAngleZ = 0.0f;
								nDistanceX = 0.0f;
								nDistanceY = 0.0f;
								nDistanceZ = -8.0f;
                        }
                        InvalidateRect (hwnd, NULL, FALSE);
                        return 0;
                }
                break;

                case WM_KEYDOWN:
                {
                        if(aFlag == 1)
                                break;

                        char chCharCode = (TCHAR) wParam;

						int ctrl_mode = 0;
						int shift_mode = 0;

						GetAsyncKeyState(VK_CONTROL);
						if(GetAsyncKeyState(VK_CONTROL))
							ctrl_mode = 1;

						GetAsyncKeyState(VK_SHIFT);
						if(GetAsyncKeyState(VK_SHIFT))
							ctrl_mode = 2;

						if(ctrl_mode == 0 && chCharCode == 0x26) // up
                        {
                                nAngleX += ROT_STEP;
                                if (nAngleX >= 360)
                                        nAngleX -= 360;
                        }

						if(ctrl_mode == 1 && chCharCode == 0x26) // up
                        {
							nDistanceZ += MOVE_STEP;
						}

						if(ctrl_mode == 2 && chCharCode == 0x26) // up
                        {
							nDistanceY += MOVE_STEP;
						}

                        if(ctrl_mode == 0 && chCharCode == 0x28) // down
                        {
							nAngleX -= ROT_STEP;
							if (nAngleX <= 0)
								nAngleX += 360;
                        }

						if(ctrl_mode == 1 && chCharCode == 0x28) // down
                        {
							nDistanceZ -= MOVE_STEP;
						}

						if(ctrl_mode == 2 && chCharCode == 0x28) // down
                        {
							nDistanceY -= MOVE_STEP;
						}

						if(ctrl_mode == 0 && chCharCode == 0x27) // right
                        {
							nAngleZ += ROT_STEP;
                            if (nAngleZ >= 360)
								nAngleZ -= 360;
                        }

						if(ctrl_mode == 1 && chCharCode == 0x27) // right
                        {
							nDistanceX += MOVE_STEP;
						}

						if(ctrl_mode == 0 && chCharCode == 0x25) // left
                        {
							nAngleZ -= ROT_STEP;
                            if (nAngleZ <= 0)
								nAngleZ += 360;
                        }

						if(ctrl_mode == 1 && chCharCode == 0x25) // right
                        {
							nDistanceX -= MOVE_STEP;
						}

						if(chCharCode == 0x6B) // +
                        {
                                nSize += 0.05f;
                        }

						if(chCharCode == 0x6D) // -
                        {
                                nSize -= 0.05f;
                                if(nSize < 0.0)
                                        nSize = 0.0;
                        }

                        InvalidateRect (hwnd, NULL, FALSE);
                        return 0;
                }
                break;

    }
    return DefWindowProc (hwnd, msg, wParam, lParam);
}

static void SetupLights( void )
{
        int i;

        GLfloat  l_ambientLight[] = { 0.5f, 0.5f, 0.5f, 1.0f };
        GLfloat  l_diffuseLight[] = { 0.9f, 0.9f, 0.9f, 1.0f };
        GLfloat  l_specular[] = { 1.0f, 1.0f, 1.0f, 1.0f};
        GLfloat  l_position[] = { 1.0f, 1.0f, 1.0f};
        GLfloat  l_direction[] = { 0.0f, 0.0f, -1.0f};

		glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE,l_ambientLight);
		glLightModelfv(GL_LIGHT_MODEL_AMBIENT ,l_ambientLight);

        /* Light Setup */
        for(i= 0; i < MAX_LIGHTS;i++)
        {
			l_diffuseLight[0] = lights[i].color[0];
            l_diffuseLight[1] = lights[i].color[1];
            l_diffuseLight[2] = lights[i].color[2];
			l_ambientLight[0] = l_diffuseLight[0] * 0.8f;
			l_ambientLight[1] = l_diffuseLight[1] * 0.8f;
            l_ambientLight[2] = l_diffuseLight[2] * 0.8f;

			l_specular[0] = l_diffuseLight[0] * 1.1f;
            l_specular[1] = l_diffuseLight[1] * 1.1f;
            l_specular[2] = l_diffuseLight[2] * 1.1f;

            l_position[0] = lights[i].position[0];
            l_position[1] = lights[i].position[1];
            l_position[2] = lights[i].position[2];

            glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE,l_ambientLight);

            glLightfv (GL_LIGHT0 + i, GL_AMBIENT, l_ambientLight);
            glLightfv (GL_LIGHT0 + i, GL_DIFFUSE, l_diffuseLight);
            glLightfv (GL_LIGHT0 + i, GL_SPECULAR, l_specular);
            glLightfv (GL_LIGHT0 + i, GL_POSITION,l_position);

            if(lights[i].isSpot)
            {
				l_direction[0] = lights[i].direction[0];
                l_direction[1] = lights[i].direction[1];
                l_direction[2] = lights[i].direction[2];

                glLightf (GL_LIGHT0 + i,GL_SPOT_CUTOFF, lights[i].radius);
                glLightf (GL_LIGHT0 + i,GL_SPOT_EXPONENT, lights[i].falloff);
                glLightfv (GL_LIGHT0 + i,GL_SPOT_DIRECTION, l_direction);

			}
            else
            {
				l_direction[0] = 0.0f;
                l_direction[1] = 0.0f;
                l_direction[2] = -1.0f;

                glLightf(GL_LIGHT0 + i,GL_SPOT_CUTOFF, 180.0f);
                glLightf(GL_LIGHT0 + i,GL_SPOT_EXPONENT, 0.0f);
                glLightfv (GL_LIGHT0 + i,GL_SPOT_DIRECTION, l_direction);
            }

            glEnable (GL_LIGHT0 + i);
        }
        glEnable (GL_LIGHTING);
}

static void SetupModels(void)
{
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glFrontFace(GL_CCW);
//    glPolygonMode(GL_BACK, GL_LINE);

	for(int k = 0; k < MAX_OBJECTS;k++)
    {
		SelectMaterial(objects[k].mat_ref);

		glBegin (GL_TRIANGLES);

		int *faces = objects[k].face_indicies;
		float *v = objects[k].verticies;
		float *n = objects[k].normals;
		float *t = objects[k].textures;

		for(int i = 0;i < objects[k].face_number;i++)
		{
			for(int j = 0;j < 3;j++)
			{
				if(n)
				{
					int ni = faces[i * 9 + j + 3];
					glNormal3f  (n[ni * 3 + 0], n[ni * 3 + 1], n[ni * 3 + 2]);
				}

                if(t)
                {
					int ti = faces[i * 9 + j + 6];
                    glTexCoord2f(t[ti * 3 + 0], t[ti * 3 + 1]);
				}

				int vi = faces[i * 9 + j + 0];
                glVertex3f  (v[vi * 3 + 0], v[vi * 3 + 1], v[vi * 3 + 2]);
			}
			face_number++;
		}
        glEnd ();
	}
}

static void Resize(void)
{
	if(glnHeight == 0)
		glnHeight = 1;

    glViewport (0, 0, glnWidth, glnHeight);

    glMatrixMode (GL_PROJECTION);
    glLoadIdentity ();
	gldAspect = (GLdouble) glnWidth / (GLdouble) glnHeight;
    gluPerspective (30.0,           // Field-of-view angle
                    gldAspect,      // Aspect ratio of viewing volume
                    0.1,           // Distance to near clipping plane
                    100.0);                   // Distance to far clipping plane

}

static void DrawScene (HDC hdc,
						float nAngleX,
						float nAngleY,
						float nAngleZ,
						float nDistanceX,
						float nDistanceY,
						float nDistanceZ)
{
    //
    // Clear the color and depth buffers.
    //
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    //
    // Define the modelview transformation.
    //
	glMatrixMode (GL_MODELVIEW);
    glLoadIdentity ();
    glTranslatef (nDistanceX, nDistanceY, nDistanceZ);
    glRotatef ( -75.0f, 1.0f, 0.0f, 0.0f);
    glRotatef (nAngleX, 1.0f, 0.0f, 0.0f);
    glRotatef (nAngleY, 0.0f, 1.0f, 0.0f);
    glRotatef (nAngleZ, 0.0f, 0.0f, 1.0f);
    glScalef (nSize+1.5f,nSize+1.5f,nSize+1.5f);
    glRotatef (30.0f, 0.0f, 0.0f, 1.0f);

	if(glIsList(LIGHT_LIST))
		glCallList(LIGHT_LIST);

	if(glIsList(POLYGON_LIST))
		glCallList(POLYGON_LIST);

	glFlush();

    SwapBuffers (hdc);
};

static void InitializeRC (void)
{
	int i;

    glClearColor(background[0], background[1], background[2], background[0] );
	glClearDepth(1.0f);									// Depth Buffer Setup
	glEnable (GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);								// The Type Of Depth Testing To Do
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	// Really Nice Perspective Calculations
	glEnable (GL_NORMALIZE);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glShadeModel(GL_SMOOTH);
//    glShadeModel(GL_FLAT);
//    glDisable(GL_CULL_FACE);
    glEnable(GL_CULL_FACE);

	for(i=0;i<MAX_TEXTUREMAPS;i++)
    {
		GLuint texture_name;
		glGenTextures(1,&texture_name);
		texture_maps[i].id=texture_name;
        glBindTexture(GL_TEXTURE_2D,texture_name);
        LoadTexture(texture_maps[i].name);
    }

	glNewList(LIGHT_LIST, GL_COMPILE);
	SetupLights();
	glEndList();

	glNewList(POLYGON_LIST, GL_COMPILE);
	SetupModels();
	glEndList();

	glClearColor(background[0], background[1], background[2], background[0] );
}

static void SetDCPixelFormat (HWND hwnd,HDC hdc)
{
    HANDLE hHeap;
    int nColors, i;
    LPLOGPALETTE lpPalette;
    BYTE byRedMask, byGreenMask, byBlueMask;

    static PIXELFORMATDESCRIPTOR pfd = {
        sizeof (PIXELFORMATDESCRIPTOR),             // Size of this structure
        1,                                          // Version number
        PFD_DRAW_TO_WINDOW |                        // Flags
        PFD_SUPPORT_OPENGL |
        PFD_GENERIC_ACCELERATED|
        PFD_DOUBLEBUFFER,
        PFD_TYPE_RGBA,                              // RGBA pixel values
        24,											// 24-bit color
        0, 0, 0, 0, 0, 0,                           // Don't care about these
        0, 0,                                       // No alpha buffer
        0, 0, 0, 0, 0,                              // No accumulation buffer
        32,											// 32-bit depth buffer
        0,                                          // No stencil buffer
        0,                                          // No auxiliary buffers
        PFD_MAIN_PLANE,                             // Layer type
        0,                                          // Reserved (must be 0)
        0, 0, 0                                     // No layer masks
    };

    int nPixelFormat;

    nPixelFormat = ChoosePixelFormat (hdc, &pfd);
    SetPixelFormat (hdc, nPixelFormat, &pfd);

    if (pfd.dwFlags & PFD_NEED_PALETTE)
	{
        nColors = 1 << pfd.cColorBits;
        hHeap = GetProcessHeap ();

        (LPLOGPALETTE) lpPalette = (LPLOGPALETTE)HeapAlloc (hHeap, 0,
            sizeof (LOGPALETTE) + (nColors * sizeof (PALETTEENTRY)));

        lpPalette->palVersion = 0x300;
        lpPalette->palNumEntries = nColors;

        byRedMask = (1 << pfd.cRedBits) - 1;
        byGreenMask = (1 << pfd.cGreenBits) - 1;
        byBlueMask = (1 << pfd.cBlueBits) - 1;

        for (i=0; i<nColors; i++) {
            lpPalette->palPalEntry[i].peRed =
                (((i >> pfd.cRedShift) & byRedMask) * 255) / byRedMask;
            lpPalette->palPalEntry[i].peGreen =
                (((i >> pfd.cGreenShift) & byGreenMask) * 255) / byGreenMask;
            lpPalette->palPalEntry[i].peBlue =
                (((i >> pfd.cBlueShift) & byBlueMask) * 255) / byBlueMask;
            lpPalette->palPalEntry[i].peFlags = 0;
        }

        hPalette = CreatePalette (lpPalette);
        HeapFree (hHeap, 0, lpPalette);

        if (hPalette != NULL)
		{
            SelectPalette (hdc, hPalette, FALSE);
            RealizePalette (hdc);
        }
    }
}

static DEVMODE currentdevMode;
static BOOL ChangeResolution(LONG w, LONG h, LONG bitdepth, int mode)
{
	DEVMODE devMode;
	LONG   modeExist;
	LONG   modeSwitch;
	LONG   i;
	char    buf[256];

	if(mode == 1)
	{
		modeSwitch = ChangeDisplaySettings(&currentdevMode, CDS_FULLSCREEN);
		if (modeSwitch==DISP_CHANGE_SUCCESSFUL)
			return TRUE;
	}

	modeExist = EnumDisplaySettings(NULL, ENUM_CURRENT_SETTINGS, &currentdevMode);
	if (modeExist)
	{
		sprintf(buf, "Current Settings %ldx%ldx%ld-bit color \n",
							currentdevMode.dmPelsWidth,
							currentdevMode.dmPelsHeight,
							currentdevMode.dmBitsPerPel
                        );
	}

  for (i=0; ;i++)
  {
    modeExist = EnumDisplaySettings(NULL, i, &devMode);

    if ( ((LONG)devMode.dmBitsPerPel == bitdepth)
       &&((LONG)devMode.dmPelsWidth == w)
       &&((LONG)devMode.dmPelsHeight == h) )
    {
      modeSwitch = ChangeDisplaySettings(&devMode, CDS_FULLSCREEN);
          if (modeSwitch==DISP_CHANGE_SUCCESSFUL)
                  return TRUE;
    }
  }
  return 0;
}

static BOOL LoadDIB(char*file,DIB2D*dib)
{
        HANDLE hfile=CreateFile(file,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING ,FILE_ATTRIBUTE_NORMAL,0);
        if(hfile!=INVALID_HANDLE_VALUE)
    {
                DWORD readed;
                int size=GetFileSize(hfile,0);
                if(size>sizeof(BITMAPFILEHEADER))
                {
                        BITMAPFILEHEADER bmfh;
                        ReadFile(hfile,&bmfh,sizeof(BITMAPFILEHEADER),&readed,0);
                        if((readed==sizeof(BITMAPFILEHEADER)) && (bmfh.bfType=='MB'))
                        {
                                dib->Info=(BITMAPINFOHEADER*)(new BYTE[size-sizeof(BITMAPFILEHEADER)]);
                                ReadFile(hfile,dib->Info,size-sizeof(BITMAPFILEHEADER),&readed,0);
                                dib->bits=(BYTE*)(dib->Info+1);

                                if(dib->Info->biBitCount==8)
                                {
                                        dib->palette=(RGBQUAD*)dib->bits;
                                        if(dib->Info->biClrUsed)
                                                dib->bits+=dib->Info->biClrUsed*4;
                                        else
                                                dib->bits+=1024;
                                }
                                else
                                {
                                        dib->palette=NULL;
                                }
                        }
                }
                CloseHandle(hfile);
                return TRUE;
        }
        return FALSE;
}

static long ScanBytes(int pixWidth, int bitsPixel)
{
  return (((long)pixWidth*bitsPixel+31) / 32) * 4;
}

static BOOL  ScaleImage(DIB2D&dib,GLTXTLOAD&p)
{
        GLint glMaxTexDim;     // OpenGL maximum texture dimension
        GLint XDMaxTexDim=512; // user maximum texture dimension
        GLint minsize =2;
        double xPow2, yPow2;
        int ixPow2, iyPow2;
        int xSize2, ySize2;
        GLint m_iWidth=dib.Info->biWidth;
        GLint m_iHeight=dib.Info->biHeight;
        glGetIntegerv(GL_MAX_TEXTURE_SIZE, &glMaxTexDim);

        glMaxTexDim = min(XDMaxTexDim, glMaxTexDim);

        if (m_iWidth <= glMaxTexDim)
                xPow2 = log((double)m_iWidth) / log(2.0);
        else
                xPow2 = log((double)glMaxTexDim) / log(2.0);

        if (m_iHeight <= glMaxTexDim)
                yPow2 = log((double)m_iHeight) / log(2.0);
        else
                yPow2 = log((double)glMaxTexDim) / log(2.0);

        ixPow2 = (int)xPow2;
        iyPow2 = (int)yPow2;

        if (xPow2 != (double)ixPow2)
                ixPow2++;
        if (yPow2 != (double)iyPow2)
                iyPow2++;

        xSize2 = 1 << ixPow2;
        ySize2 = 1 << iyPow2;

        if(xSize2<minsize)
                xSize2=minsize;
        if(ySize2<minsize)
                ySize2=minsize;

        if(((xSize2==m_iWidth) && (ySize2==m_iHeight)))
        {
                if(dib.Info->biBitCount==24)
                {
                        p.format=GL_BGR_EXT;
                        p.perpixel=3;
                        return FALSE;
                }

                if(dib.Info->biBitCount==32)
                {
                        p.format=GL_BGRA_EXT;
                        p.perpixel=4;
                        return FALSE;
                }
        }

        BYTE *bits=(BYTE *)dib.bits;
        if(dib.Info->biBitCount==8)
        {

            // convert to TRUECOLOR
                int _perline=ScanBytes(8,m_iWidth);
                int perline=ScanBytes(24,m_iWidth);
                bits= new BYTE[perline*m_iHeight * sizeof(BYTE)];
                for(int y=0;y<m_iHeight;y++)
                {
                        BYTE *_b=((BYTE *)dib.bits)+y*_perline;
                        BYTE *b=bits+y*perline;
                        for(int x=0;x<m_iWidth;x++)
                        {
                                RGBQUAD _p=dib.palette[*_b];
                                _b++;
                                *b=_p.rgbBlue;b++;
                                *b=_p.rgbGreen;b++;
                                *b=_p.rgbRed;b++;
                        }
                }
        }

        BOOL isAlpha=(dib.Info->biBitCount==32);
        int _mem_size=xSize2 * ySize2 *  sizeof(BYTE);
        if(isAlpha)
        {
                _mem_size*=4;
        p.perpixel=4;
        p.format=GL_BGRA_EXT;
        }
        else
        {
                _mem_size*=3;
        p.perpixel=3;
        p.format=GL_BGR_EXT;
    }

        BYTE *pData = (BYTE*)new BYTE[_mem_size];
        if (!pData)
                return FALSE;

        if(isAlpha)
                gluScaleImage(GL_BGRA_EXT, m_iWidth, m_iHeight,
                                        GL_UNSIGNED_BYTE, bits,
                                        xSize2, ySize2, GL_UNSIGNED_BYTE, pData);
        else
                gluScaleImage(GL_RGB, m_iWidth, m_iHeight,
                                        GL_UNSIGNED_BYTE, bits,
                                        xSize2, ySize2, GL_UNSIGNED_BYTE, pData);


        if(bits!=dib.bits)
                delete bits;

        m_iWidth        = xSize2;
        m_iHeight       = ySize2;
        p.Width         = m_iWidth;
        p.Height        = m_iHeight;
        p.bits          = pData;

        return TRUE ;
}


static void LoadTexture(char*filename)
{
        DIB2D dib;
        GLTXTLOAD load;
        if(LoadDIB(filename,&dib))
        {

                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

                if(ScaleImage(dib,load))
                {
                        glTexImage2D(GL_TEXTURE_2D,0,load.perpixel,
                                                load.Width,load.Height,0,
                                                load.format,GL_UNSIGNED_BYTE,
                                                load.bits);
                        delete load.bits;
                }
                else
                {

                        glTexImage2D(GL_TEXTURE_2D,0,load.perpixel,
                                                dib.Info->biWidth,dib.Info->biHeight,
                                                0,load.format,GL_UNSIGNED_BYTE,dib.bits);
                }
                delete dib.Info;
        }
};

static void SetMaterial(GLenum mat_property, GLfloat *tmp, GLfloat alpha)
{
	GLfloat data[4];
    data[0] = tmp[0];
    data[1] = tmp[1];
    data[2] = tmp[2];
    data[3] = alpha;
    glMaterialfv (GL_FRONT_AND_BACK, mat_property, data);
    glMaterialf (GL_FRONT_AND_BACK, GL_SHININESS,100);
}

static void SelectMaterial(int i)
{
	GLfloat alpha = materials[i].alpha;

    alpha = 1.0f;
	SetMaterial (GL_AMBIENT, materials[i].ambient,alpha);
    SetMaterial (GL_DIFFUSE, materials[i].diffuse,alpha);
    SetMaterial (GL_SPECULAR, materials[i].specular,alpha);
    SetMaterial (GL_EMISSION, materials[i].emission,alpha);

    if(MAX_TEXTUREMAPS && (materials[i].texture > (-1)))
    {
		glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D,texture_maps[materials[i].texture].id);
	}
    else
    {
		glDisable(GL_TEXTURE_2D);
	}
}
